__author__ = "Marc Hannappel <salandora@gmail.com>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2017 The OctoPrint Project - Released under terms of the AGPLv3 License"

from collections import OrderedDict, defaultdict
from functools import wraps

from flask import abort, g
from flask_babel import gettext

from octoprint.access import ADMIN_GROUP, READONLY_GROUP, USER_GROUP
from octoprint.vendor.flask_principal import Need, Permission, PermissionDenied, RoleNeed


class OctoPrintPermission(Permission):
    @classmethod
    def convert_needs_to_dict(cls, needs):
        ret_needs = defaultdict(list)
        for need in needs:
            if need.value not in ret_needs[need.method]:
                ret_needs[need.method].append(need.value)
        return ret_needs

    @classmethod
    def convert_to_needs(cls, needs):
        result = []
        for need in needs:
            if isinstance(need, Need):
                result.append(need)
            elif isinstance(need, Permission):
                result += need.needs
            elif isinstance(need, str):
                result.append(RoleNeed(need))
        return result

    def __init__(self, name, description, *needs, **kwargs):
        self._name = name
        self._description = description
        self._dangerous = kwargs.pop("dangerous", False)
        self._default_groups = kwargs.pop("default_groups", [])

        self._key = None

        Permission.__init__(self, *self.convert_to_needs(needs))

    def as_dict(self):
        return {
            "key": self.key,
            "name": self.get_name(),
            "dangerous": self._dangerous,
            "default_groups": self._default_groups,
            "description": self.get_description(),
            "needs": self.convert_needs_to_dict(self.needs),
        }

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        self._key = value

    @property
    def dangerous(self):
        return self._dangerous

    @property
    def default_groups(self):
        return self._default_groups

    def get_name(self):
        return self._name

    def get_description(self):
        return self._description

    def allows(self, identity):
        """Whether the identity can access this permission.
        Overridden from Permission.allows to make sure the Identity provides ALL
        required needs instead of ANY required need.

        :param identity: The identity
        """
        if self.needs and len(self.needs.intersection(identity.provides)) != len(
            self.needs
        ):
            return False

        if self.excludes and self.excludes.intersection(identity.provides):
            return False

        return True

    def union(self, other):
        """Create a new OctoPrintPermission with the requirements of the union of this
        and other.

        :param other: The other permission
        """
        p = self.__class__(self._name, self._description, *self.needs.union(other.needs))
        p.excludes.update(self.excludes.union(other.excludes))
        return p

    def difference(self, other):
        """Create a new OctoPrintPermission consisting of requirements in this
        permission and not in the other.
        """

        p = self.__class__(
            self._name, self._description, *self.needs.difference(other.needs)
        )
        p.excludes.update(self.excludes.difference(other.excludes))
        return p

    def __repr__(self):
        return "{}({!r}, {!r}, {})".format(
            self.__class__.__name__,
            self.get_name(),
            self.get_description(),
            ", ".join(map(repr, self.needs)),
        )

    def __hash__(self):
        return self.get_name().__hash__()

    def __eq__(self, other):
        return (
            isinstance(other, OctoPrintPermission) and other.get_name() == self.get_name()
        )


class PluginOctoPrintPermission(OctoPrintPermission):
    def __init__(self, *args, **kwargs):
        self.plugin = kwargs.pop("plugin", None)
        OctoPrintPermission.__init__(self, *args, **kwargs)

    def as_dict(self):
        result = OctoPrintPermission.as_dict(self)
        result["plugin"] = self.plugin
        return result


class PluginIdentityContext:
    """Identity context for not initialized Permissions

    Needed to support @Permissions.PLUGIN_X_Y.require()

    Will search the permission when needed
    """

    def __init__(self, key, http_exception=None):
        self.key = key
        self.http_exception = http_exception
        """The permission of this principal
        """

    @property
    def identity(self):
        """The identity of this principal"""
        return g.identity

    def can(self):
        """Whether the identity has access to the permission"""
        permission = getattr(Permissions, self.key)
        if permission is None or isinstance(permission, PluginPermissionDecorator):
            raise UnknownPermission(self.key)

        return permission.can()

    def __call__(self, f):
        @wraps(f)
        def _decorated(*args, **kw):
            with self:
                rv = f(*args, **kw)
            return rv

        return _decorated

    def __enter__(self):
        permission = getattr(Permissions, self.key)
        if permission is None or isinstance(permission, PluginPermissionDecorator):
            raise UnknownPermission(self.key)

        # check the permission here
        if not permission.can():
            if self.http_exception:
                abort(self.http_exception, permission)
            raise PermissionDenied(permission)

    def __exit__(self, *args):
        return False


class PluginPermissionDecorator(Permission):
    """Decorator Class for not initialized Permissions

    Needed to support @Permissions.PLUGIN_X_Y.require()
    """

    def __init__(self, key):
        self.key = key

    def require(self, http_exception=None):
        return PluginIdentityContext(self.key, http_exception)


class PermissionsMetaClass(type):
    permissions = OrderedDict()

    def __new__(mcs, name, bases, args):
        cls = type.__new__(mcs, name, bases, args)

        for key, value in args.items():
            if isinstance(value, OctoPrintPermission):
                value.key = key
                mcs.permissions[key] = value
                delattr(cls, key)

        return cls

    def __setattr__(cls, key, value):
        if isinstance(value, OctoPrintPermission):
            if key in cls.permissions:
                raise PermissionAlreadyExists(key)
            value.key = key
            cls.permissions[key] = value

    def __getattr__(cls, key):
        permission = cls.permissions.get(key)

        if key.startswith("PLUGIN_") and permission is None:
            return PluginPermissionDecorator(key)

        return permission

    def all(cls):
        return list(cls.permissions.values())

    def filter(cls, cb):
        return list(filter(cb, cls.all()))

    def find(cls, p, filter=None):
        key = None
        if isinstance(p, OctoPrintPermission):
            key = p.key
        elif isinstance(p, dict):
            key = p.get("key")
        elif isinstance(p, str):
            key = p

        if key is None:
            return None

        return cls.match(lambda p: p.key == key, filter=filter)

    def match(cls, match, filter=None):
        if callable(filter):
            permissions = cls.filter(filter)
        else:
            permissions = cls.all()

        for permission in permissions:
            if match(permission):
                return permission

        return None


class Permissions(metaclass=PermissionsMetaClass):

    # Special permission
    ADMIN = OctoPrintPermission(
        "Admin",
        gettext("Admin is allowed to do everything"),
        RoleNeed("admin"),
        dangerous=True,
        default_groups=[ADMIN_GROUP],
    )

    STATUS = OctoPrintPermission(
        "Status",
        gettext(
            "Allows to gather basic status information, e.g. job progress, "
            "printer state, temperatures, ... Mandatory for the default UI "
            "to work"
        ),
        RoleNeed("status"),
        default_groups=[USER_GROUP, READONLY_GROUP],
    )

    CONNECTION = OctoPrintPermission(
        "Connection",
        gettext("Allows to connect to and disconnect from a printer"),
        RoleNeed("connection"),
        default_groups=[USER_GROUP],
    )

    WEBCAM = OctoPrintPermission(
        "Webcam",
        gettext("Allows to watch the webcam stream"),
        RoleNeed("webcam"),
        default_groups=[USER_GROUP, READONLY_GROUP],
    )

    SYSTEM = OctoPrintPermission(
        "System",
        gettext(
            "Allows to run system commands, e.g. restart OctoPrint, "
            "shutdown or reboot the system, and to retrieve system and usage information"
        ),
        RoleNeed("system"),
        dangerous=True,
    )

    FILES_LIST = OctoPrintPermission(
        "File List",
        gettext(
            "Allows to retrieve a list of all uploaded files and folders, including"
            "their metadata (e.g. date, file size, analysis results, ...)"
        ),
        RoleNeed("files_list"),
        default_groups=[USER_GROUP, READONLY_GROUP],
    )
    FILES_UPLOAD = OctoPrintPermission(
        "File Upload",
        gettext(
            "Allows users to upload new files, create new folders and copy existing ones. If "
            "the File Delete permission is also set, File Upload also allows "
            "moving files and folders."
        ),
        RoleNeed("files_upload"),
        default_groups=[USER_GROUP],
    )
    FILES_DOWNLOAD = OctoPrintPermission(
        "File Download",
        gettext(
            "Allows users to download files. The GCODE viewer is "
            "affected by this as well."
        ),
        RoleNeed("files_download"),
        default_groups=[USER_GROUP, READONLY_GROUP],
    )
    FILES_DELETE = OctoPrintPermission(
        "File Delete",
        gettext(
            "Allows users to delete files and folders. If the File Upload permission is "
            "also set, File Delete also allows moving files and folders."
        ),
        RoleNeed("files_delete"),
        default_groups=[USER_GROUP],
    )
    FILES_SELECT = OctoPrintPermission(
        "File Select",
        gettext("Allows to select a file for printing"),
        RoleNeed("files_select"),
        default_groups=[USER_GROUP],
    )

    PRINT = OctoPrintPermission(
        "Print",
        gettext("Allows to start, pause and cancel a print job"),
        RoleNeed("print"),
        default_groups=[USER_GROUP],
    )

    GCODE_VIEWER = OctoPrintPermission(
        "GCODE viewer",
        gettext(
            'Allows access to the GCODE viewer if the "File Download"'
            "permission is also set."
        ),
        RoleNeed("gcodeviewer"),
        default_groups=[USER_GROUP, READONLY_GROUP],
    )

    MONITOR_TERMINAL = OctoPrintPermission(
        "Terminal",
        gettext(
            "Allows to watch the terminal tab but not to send commands "
            "to the printer from it"
        ),
        RoleNeed("monitor_terminal"),
        default_groups=[USER_GROUP, READONLY_GROUP],
    )

    CONTROL = OctoPrintPermission(
        "Control",
        gettext(
            "Allows to control of the printer by using the temperature controls,"
            "the control tab or sending commands through the terminal."
        ),
        RoleNeed("control"),
        default_groups=[USER_GROUP],
    )

    SLICE = OctoPrintPermission(
        "Slice",
        gettext("Allows to slice files"),
        RoleNeed("slice"),
        default_groups=[USER_GROUP],
    )

    TIMELAPSE_LIST = OctoPrintPermission(
        "Timelapse List",
        gettext("Allows to list timelapse videos"),
        RoleNeed("timelapse_list"),
        default_groups=[USER_GROUP, READONLY_GROUP],
    )
    TIMELAPSE_DOWNLOAD = OctoPrintPermission(
        "Timelapse Download",
        gettext("Allows to download timelapse videos"),
        RoleNeed("timelapse_download"),
        default_groups=[USER_GROUP, READONLY_GROUP],
    )
    TIMELAPSE_DELETE = OctoPrintPermission(
        "Timelapse Delete",
        gettext("Allows to delete timelapse videos and unrendered timelapses"),
        RoleNeed("timelapse_delete"),
        default_groups=[USER_GROUP],
    )
    TIMELAPSE_ADMIN = OctoPrintPermission(
        "Timelapse Admin",
        gettext(
            "Allows to change the timelapse settings and delete or "
            'render unrendered timelapses. Includes the "Timelapse List",'
            '"Timelapse Delete" and "Timelapse Download" permissions'
        ),
        RoleNeed("timelapse_admin"),
        TIMELAPSE_LIST,
        TIMELAPSE_DOWNLOAD,
        default_groups=[USER_GROUP],
    )

    SETTINGS_READ = OctoPrintPermission(
        "Settings Access",
        gettext(
            "Allows to read non sensitive settings. Mandatory for the "
            "default UI to work."
        ),
        RoleNeed("settings_read"),
        default_groups=[USER_GROUP, READONLY_GROUP],
    )
    SETTINGS = OctoPrintPermission(
        "Settings Admin",
        gettext("Allows to manage settings and also to read sensitive settings"),
        RoleNeed("settings"),
        dangerous=True,
    )


class PermissionAlreadyExists(Exception):
    def __init__(self, permission):
        Exception.__init__(self, "Permission %s already exists" % permission)


class UnknownPermission(Exception):
    def __init__(self, permissionname):
        Exception.__init__(self, "Unknown permission: %s" % permissionname)



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2015 The OctoPrint Project - Released under terms of the AGPLv3 License"


import logging
import os
from concurrent.futures import ThreadPoolExecutor, wait

import click

from octoprint import FatalStartupError, init_pluginsystem, init_settings
from octoprint.cli import get_ctx_obj_option, standard_options
from octoprint.timelapse import create_thumbnail_path, valid_timelapse
from octoprint.util import is_hidden_path

click.disable_unicode_literals_warning = True


@click.group()
@click.pass_context
def cli(ctx):
    """Basic config manipulation."""
    logging.basicConfig(
        level=logging.DEBUG
        if get_ctx_obj_option(ctx, "verbosity", 0) > 0
        else logging.WARN
    )
    try:
        ctx.obj.settings = init_settings(
            get_ctx_obj_option(ctx, "basedir", None),
            get_ctx_obj_option(ctx, "configfile", None),
            overlays=get_ctx_obj_option(ctx, "overlays", None),
        )
        ctx.obj.plugin_manager = init_pluginsystem(
            ctx.obj.settings, safe_mode=get_ctx_obj_option(ctx, "safe_mode", False)
        )
    except FatalStartupError as e:
        click.echo(str(e), err=True)
        click.echo("There was a fatal error initializing the settings manager.", err=True)
        ctx.exit(-1)


@cli.command(name="create_thumbnails")
@standard_options(hidden=True)
@click.option(
    "--missing",
    is_flag=True,
    help="Create thumbnails for all timelapses that don't yet have one",
)
@click.option(
    "--processes",
    type=int,
    help="Number of processes to use for creating thumbnails",
    default=1,
)
@click.argument("paths", nargs=-1, type=click.Path())
@click.pass_context
def extract_thumbnails(ctx, missing, processes, paths):
    """Extract missing thumbnails from timelapses."""

    if not paths and not missing:
        click.echo(
            "No paths specified and no --missing flag either, nothing to do.", err=True
        )
        ctx.exit(-1)

    if missing:
        paths = []
        timelapse_folder = ctx.obj.settings.getBaseFolder("timelapse")
        for entry in os.scandir(timelapse_folder):
            if is_hidden_path(entry.path) or not valid_timelapse(entry.path):
                continue

            thumb = create_thumbnail_path(entry.path)
            if not os.path.isfile(thumb):
                paths.append(entry.path)

    ffmpeg = ctx.obj.settings.get(["webcam", "ffmpeg"])

    click.echo(f"Creating thumbnails for {len(paths)} timelapses")
    executor = ThreadPoolExecutor(max_workers=processes)
    futures = [executor.submit(_create_thumbnail, ffmpeg, path) for path in paths]
    wait(futures)
    click.echo("Done!")


def _create_thumbnail(ffmpeg, path):
    from octoprint.timelapse import TimelapseRenderJob

    click.echo(f"Creating thumbnail for {path}...")
    return TimelapseRenderJob._try_generate_thumbnail(ffmpeg, path)



"""CSRF double cookie implementation"""

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2022 The OctoPrint Project - Released under terms of the AGPLv3 License"

import flask

from octoprint.server.util.flask import OctoPrintFlaskResponse

exempt_views = set()


def _get_view_location(view):
    if isinstance(view, str):
        return view
    else:
        return f"{view.__module__}.{view.__name__}"


def add_exempt_view(view):
    exempt_views.add(_get_view_location(view))


def is_exempt(view):
    if not view:
        return False

    return _get_view_location(view) in exempt_views


def generate_csrf_token():
    import hashlib
    import os

    from itsdangerous import URLSafeTimedSerializer

    value = hashlib.sha1(os.urandom(64)).hexdigest()

    secret = flask.current_app.config["SECRET_KEY"]
    s = URLSafeTimedSerializer(secret, salt="octoprint-csrf-token")

    return s.dumps(value)


def validate_csrf_tokens(cookie, header):
    import hmac

    from itsdangerous import URLSafeTimedSerializer

    if cookie is None or header is None:
        return False

    secret = flask.current_app.config["SECRET_KEY"]
    s = URLSafeTimedSerializer(secret, salt="octoprint-csrf-token")

    try:
        # TODO: set max age for values, once we have a REST based heartbeat
        # that takes care of regular cookie refresh
        value_cookie = s.loads(cookie)
        value_header = s.loads(header)
        return hmac.compare_digest(value_cookie, value_header)
    except Exception:
        return False


def add_csrf_cookie(response):
    if not isinstance(response, OctoPrintFlaskResponse):
        return

    token = generate_csrf_token()
    response.set_cookie("csrf_token", token, httponly=False)
    return response


def validate_csrf_request(request):
    if request.method in ["GET", "HEAD", "OPTIONS"]:
        # Irrelevant method for CSRF, bypass
        return

    if getattr(flask.g, "login_via_apikey", False):
        # API key authorization, bypass
        return

    if is_exempt(request.endpoint):
        # marked as exempt, bypass
        return

    cookie = request.cookies.get("csrf_token", None)
    header = request.headers.get("X-CSRF-Token", None)

    if not validate_csrf_tokens(cookie, header):
        flask.abort(400, "CSRF validation failed")


def csrf_exempt(view):
    add_exempt_view(view)
    return view



"""
This module provides a bunch of utility methods and helpers for version handling.
"""

__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"

import logging

import pkg_resources

from octoprint import __version__


def get_octoprint_version_string():
    return __version__


def get_octoprint_version(cut=None, **kwargs):
    octoprint_version_string = normalize_version(get_octoprint_version_string())
    return get_comparable_version(octoprint_version_string, cut=cut, **kwargs)


def is_released_octoprint_version(version=None):
    if version is None:
        version = get_octoprint_version()
    return is_release(version)


def is_stable_octoprint_version(version=None):
    if version is None:
        version = get_octoprint_version()
    return is_stable(version)


def is_octoprint_compatible(*compatibility_entries, **kwargs):
    """
    Tests if the current ``octoprint_version`` is compatible to any of the provided ``compatibility_entries``.

    Arguments:
            compatibility_entries (str): compatibility string(s) to test against, result will be `True` if any match
                    is found
            octoprint_version (tuple or SetuptoolsVersion): optional OctoPrint version to match against, if not current
                    base version will be determined via :func:`get_octoprint_version`.

    Returns:
            (bool) ``True`` if any of the provided compatibility entries matches or there are no entries, else ``False``
    """

    logger = logging.getLogger(__name__)

    if not compatibility_entries:
        return True

    octoprint_version = kwargs.get("octoprint_version")
    if octoprint_version is None:
        octoprint_version = get_octoprint_version(base=True)

    for octo_compat in compatibility_entries:
        try:
            if not any(
                octo_compat.startswith(c)
                for c in ("<", "<=", "!=", "==", ">=", ">", "~=", "===")
            ):
                octo_compat = f">={octo_compat}"

            s = pkg_resources.Requirement.parse("OctoPrint" + octo_compat)
            if octoprint_version in s:
                break
        except Exception:
            logger.exception(
                "Something is wrong with this compatibility string for OctoPrint: {}".format(
                    octo_compat
                )
            )
    else:
        return False

    return True


def get_python_version_string():
    from platform import python_version

    version_string = normalize_version(python_version())

    return version_string


def get_python_version():
    return get_comparable_version(get_python_version_string())


def is_python_compatible(compat, **kwargs):
    if not compat:
        return True

    python_version = kwargs.get("python_version")
    if python_version is None:
        python_version = get_python_version_string()

    s = pkg_resources.Requirement.parse("Python" + compat)
    return python_version in s


def get_comparable_version(version_string, cut=None, **kwargs):
    """
    Args:
        version_string: The version string for which to create a comparable version instance
        cut: optional, how many version digits to remove (e.g., cut=1 will turn 1.2.3 into 1.2).
             Defaults to ``None``, meaning no further action. Settings this to 0 will remove
             anything up to the last digit, e.g. dev or rc information.

    Returns:
        A comparable version
    """

    if "base" in kwargs and kwargs.get("base", False) and cut is None:
        cut = 0
    if cut is not None and (cut < 0 or not isinstance(cut, int)):
        raise ValueError("level must be a positive integer")

    version_string = normalize_version(version_string)
    version = pkg_resources.parse_version(version_string)

    if cut is not None:
        if isinstance(version, tuple):
            # old setuptools
            base_version = []
            for part in version:
                if part.startswith("*"):
                    break
                base_version.append(part)
            if 0 < cut < len(base_version):
                base_version = base_version[:-cut]
            base_version.append("*final")
            version = tuple(base_version)
        else:
            # new setuptools
            version = pkg_resources.parse_version(version.base_version)
            if cut is not None:
                parts = version.base_version.split(".")
                if 0 < cut < len(parts):
                    reduced = parts[:-cut]
                    version = pkg_resources.parse_version(
                        ".".join(str(x) for x in reduced)
                    )

    return version


def is_stable(version):
    """
    >>> import pkg_resources
    >>> is_stable(pkg_resources.parse_version("1.3.6rc3"))
    False
    >>> is_stable(pkg_resources.parse_version("1.3.6rc3.dev2+g1234"))
    False
    >>> is_stable(pkg_resources.parse_version("1.3.6"))
    True
    >>> is_stable(pkg_resources.parse_version("1.3.6.post1+g1234"))
    True
    >>> is_stable(pkg_resources.parse_version("1.3.6.post1.dev0+g1234"))
    False
    >>> is_stable(pkg_resources.parse_version("1.3.7.dev123+g23545"))
    False
    """

    if isinstance(version, str):
        version = get_comparable_version(version)

    if not is_release(version):
        return False

    if isinstance(version, tuple):
        return "*a" not in version and "*b" not in version and "*c" not in version
    else:
        return not version.is_prerelease


def is_release(version):
    """
    >>> import pkg_resources
    >>> is_release(pkg_resources.parse_version("1.3.6rc3"))
    True
    >>> is_release(pkg_resources.parse_version("1.3.6rc3.dev2+g1234"))
    False
    >>> is_release(pkg_resources.parse_version("1.3.6"))
    True
    >>> is_release(pkg_resources.parse_version("1.3.6.post1+g1234"))
    True
    >>> is_release(pkg_resources.parse_version("1.3.6.post1.dev0+g1234"))
    False
    >>> is_release(pkg_resources.parse_version("1.3.7.dev123+g23545"))
    False
    """

    if isinstance(version, str):
        version = get_comparable_version(version)

    if isinstance(version, tuple):
        # old setuptools
        return "*@" not in version
    else:
        # new setuptools
        return "dev" not in version.public
    pass


def is_prerelease(version):
    if isinstance(version, str):
        version = get_comparable_version(version)

    if isinstance(version, tuple):
        # old setuptools
        return any(map(lambda x: x in version, ("*a", "*b", "*c", "*rc")))
    else:
        # new setuptools
        return version.is_prerelease


def normalize_version(version):

    # Debian has the python version set to 2.7.15+ which is not PEP440 compliant (bug 914072)
    if version.endswith("+"):
        version = version[:-1]

    if version[0].lower() == "v":
        version = version[1:]

    return version.strip()



__author__ = "Gina Häußge <osd@foosel.net>"
__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2014 The OctoPrint Project - Released under terms of the AGPLv3 License"

from . import (  # noqa: F401
    pip,
    python_updater,
    single_file_plugin,
    sleep_a_bit,
    update_script,
)



__license__ = "GNU Affero General Public License http://www.gnu.org/licenses/agpl.html"
__copyright__ = "Copyright (C) 2020 The OctoPrint Project - Released under terms of the AGPLv3 License"


import datetime
import logging
import os

import click
from zipstream.ng import ZIP_DEFLATED, ZipStream

from octoprint.cli import init_platform_for_cli, standard_options

click.disable_unicode_literals_warning = True


def get_systeminfo(
    environment_detector, connectivity_checker, settings, additional_fields=None
):
    from octoprint import __version__
    from octoprint.util import dict_flatten

    if additional_fields is None:
        additional_fields = {}

    environment_detector.run_detection(notify_plugins=False)

    safe_mode_file = os.path.join(settings.getBaseFolder("data"), "last_safe_mode")
    last_safe_mode = {"date": "unknown", "reason": "unknown"}
    try:
        if os.path.exists(safe_mode_file):
            with open(safe_mode_file) as f:
                last_safe_mode["reason"] = f.readline().strip()
            last_safe_mode["date"] = (
                datetime.datetime.utcfromtimestamp(
                    os.path.getmtime(safe_mode_file)
                ).isoformat()[:19]
                + "Z"
            )
    except Exception as ex:
        logging.getLogger(__name__).error(
            "Error while retrieving last safe mode information from {}: {}".format(
                safe_mode_file, ex
            )
        )

    systeminfo = {
        "octoprint": {"version": __version__, "last_safe_mode": last_safe_mode},
        "connectivity": connectivity_checker.as_dict(),
        "env": environment_detector.environment,
        "systeminfo": {"generated": datetime.datetime.utcnow().isoformat()[:19] + "Z"},
    }

    # flatten and filter
    flattened = dict_flatten(systeminfo)
    flattened["env.python.virtualenv"] = "env.python.virtualenv" in flattened

    for k, v in additional_fields.items():
        if k not in flattened:
            flattened[k] = v

    return flattened


def get_systeminfo_bundle(systeminfo, logbase, printer=None, plugin_manager=None):
    from octoprint.util import to_bytes

    try:
        z = ZipStream(compress_type=ZIP_DEFLATED)
    except RuntimeError:
        # no zlib support
        z = ZipStream(sized=True)

    if printer and printer.is_operational():
        firmware_info = printer.firmware_info
        if firmware_info:
            # add firmware to systeminfo so it's included in systeminfo.txt
            systeminfo["printer.firmware"] = firmware_info["name"]

        # Add printer log, if available
        if hasattr(printer, "_log"):
            z.add(to_bytes("\n".join(printer._log)), arcname="terminal.txt")

    # add systeminfo
    systeminfotxt = []
    for k in sorted(systeminfo.keys()):
        systeminfotxt.append(f"{k}: {systeminfo[k]}")

    z.add(to_bytes("\n".join(systeminfotxt)), arcname="systeminfo.txt")

    # add logs
    for log in (
        "octoprint.log",
        "serial.log",
    ):
        logpath = os.path.join(logbase, log)
        if os.path.exists(logpath):
            z.add_path(logpath, arcname=log)

    # add additional bundle contents from bundled plugins
    if plugin_manager:
        for name, hook in plugin_manager.get_hooks(
            "octoprint.systeminfo.additional_bundle_files"
        ).items():
            try:
                plugin = plugin_manager.get_plugin_info(name)
                if not plugin.bundled:
                    # we only support this for bundled plugins because we don't want
                    # third party logs to blow up the bundles
                    continue

                logs = hook()

                for log, content in logs.items():
                    if isinstance(content, str):
                        # log path
                        if os.path.exists(content) and os.access(content, os.R_OK):
                            z.add_path(content, arcname=log)
                    elif callable(content):
                        # content generating callable
                        z.add(to_bytes(content()), arcname=log)
            except Exception:
                logging.getLogger(__name__).exception(
                    "Error while retrieving additional bundle contents for plugin {}".format(
                        name
                    ),
                    extra={"plugin": name},
                )

    return z


def get_systeminfo_bundle_name():
    import time

    return "octoprint-systeminfo-{}.zip".format(time.strftime("%Y%m%d%H%M%S"))


@click.group()
def cli():
    pass


@cli.command(name="systeminfo")
@standard_options()
@click.argument(
    "path",
    nargs=1,
    required=False,
    type=click.Path(writable=True, dir_okay=True, resolve_path=True),
)
@click.pass_context
def systeminfo_command(ctx, path, **kwargs):
    """Retrieves and prints the system info."""
    logging.disable(logging.ERROR)
    try:
        (
            settings,
            logger,
            safe_mode,
            event_manager,
            connectivity_checker,
            plugin_manager,
            environment_detector,
        ) = init_platform_for_cli(ctx)
    except Exception as e:
        click.echo(str(e), err=True)
        click.echo("There was a fatal error initializing the platform.", err=True)
        ctx.exit(-1)
    else:
        systeminfo = get_systeminfo(
            environment_detector, connectivity_checker, {"systeminfo.generator": "cli"}
        )

        if path:
            # create zip at path
            zipfilename = os.path.join(path, get_systeminfo_bundle_name())
            click.echo(f"Writing systeminfo bundle to {zipfilename}...")

            z = get_systeminfo_bundle(
                systeminfo, settings.getBaseFolder("logs"), plugin_manager=plugin_manager
            )
            try:
                with open(zipfilename, "wb") as f:
                    f.writelines(z)
            except Exception as e:
                click.echo(str(e), err=True)
                click.echo(f"There was an error writing to {zipfilename}.", err=True)
                ctx.exit(-1)

            click.echo("Done!")
            click.echo(zipfilename)

        else:
            # output systeminfo to console
            for k in sorted(systeminfo.keys()):
                click.echo(f"{k}: {systeminfo[k]}")
    ctx.exit(0)
